/*************************************************************************************************
 * The command line interface of web crawler
 *                                                      Copyright (C) 2004-2006 Mikio Hirabayashi
 * This file is part of Hyper Estraier.
 * Hyper Estraier is free software; you can redistribute it and/or modify it under the terms of
 * the GNU Lesser General Public License as published by the Free Software Foundation; either
 * version 2.1 of the License or any later version.  Hyper Estraier is distributed in the hope
 * that it will be useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 * License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with Hyper
 * Estraier; if not, write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA 02111-1307 USA.
 *************************************************************************************************/


#include "wavermod.h"
#include "myimage.dat"


#define NUMBUFSIZ      32                /* size of a buffer for a number */
#define URIBUFSIZ      8192              /* size of a buffer for an URI */
#define SLEEPUSEC      100000            /* sleep time in micro seconds */

typedef struct {                         /* type of structure for interaction of a URL */
  WAVER *waver;                          /* waver handle */
  char *url;                             /* URL */
  int depth;                             /* depth */
} TARGSURL;


/* global variables */
pthread_mutex_t g_mutex = PTHREAD_MUTEX_INITIALIZER;  /* global mutex */
int g_thnum = 0;                         /* number of running threads */
const char *g_progname;                  /* program name */
int g_sigterm = FALSE;                   /* flag for termination signal */
const char *g_pxhost = NULL;             /* host name of the proxy */
int g_pxport = 0;                        /* host name of the proxy */
int g_timeout = -1;                      /* timeout in seconds */
int g_inputlang = ESTLANGEN;             /* prefered language */


/* function prototypes */
int main(int argc, char **argv);
static void usage(void);
static void setsignals(void);
static void sigtermhandler(int num);
static char *dequeue(WAVER *waver, int *dp);
static void enqueuelinks(WAVER *waver, CBLIST *links, CBMAP *kwords, int depth);
static int runinit(int argc, char **argv);
static int runcrawl(int argc, char **argv);
static int rununittest(int argc, char **argv);
static int runfetch(int argc, char **argv);
static int procinit(const char *rootdir);
static int proccrawl(const char *rootdir, int cont);
static int procunittest(const char *rootdir);
static int procfetch(const char *url);
static int strtolang(const char *str);
static char *capitalize(const char *str);
static int accessthnum(int inc);
static void *geturldoc(void *args);


/* main routine */
int main(int argc, char **argv){
  const char *tmp;
  int rv;
  if((tmp = getenv("ESTDBGFD")) != NULL) dpdbgfd = atoi(tmp);
  cbstdiobin();
  g_progname = argv[0];
  g_sigterm = FALSE;
  if(!est_init_net_env()){
    log_print(LL_ERROR, "could not initialize network environment");
    exit(1);
  }
  atexit(est_free_net_env);
  if(argc < 2) usage();
  rv = 0;
  if(!strcmp(argv[1], "init")){
    rv = runinit(argc, argv);
  } else if(!strcmp(argv[1], "crawl")){
    setsignals();
    rv = runcrawl(argc, argv);
  } else if(!strcmp(argv[1], "unittest")){
    rv = rununittest(argc, argv);
  } else if(!strcmp(argv[1], "fetch")){
    rv = runfetch(argc, argv);
  } else {
    usage();
  }
  return rv;
}


/* print the usage and exit */
static void usage(void){
  fprintf(stderr, "%s: command line interface of web crawler\n", g_progname);
  fprintf(stderr, "\n");
  fprintf(stderr, "usage:\n");
  fprintf(stderr, "  %s init rootdir\n", g_progname);
  fprintf(stderr, "  %s crawl [-cont] rootdir\n", g_progname);
  fprintf(stderr, "  %s unittest rootdir\n", g_progname);
  fprintf(stderr, "  %s fetch [-proxy host port] [-tout num] [-il lang] url\n", g_progname);
  fprintf(stderr, "\n");
  exit(1);
}


/* set signal handlers */
static void setsignals(void){
  est_signal(1, sigtermhandler);
  est_signal(2, sigtermhandler);
  est_signal(3, sigtermhandler);
  est_signal(13, SIG_IGN);
  est_signal(15, sigtermhandler);
  g_sigterm = FALSE;
}


/* handler of termination signal */
static void sigtermhandler(int num){
  g_sigterm = TRUE;
  log_print(LL_INFO, "the termination signal %d catched", num);
}


/* dequeue a URL from the priority queue */
static char *dequeue(WAVER *waver, int *dp){
  char *buf, *rp, *rv;
  if(time(NULL) > waver->stime + waver->period) return NULL;
  if(pthread_mutex_lock(&g_mutex) != 0) return NULL;
  if(est_mtdb_doc_num(waver->index) > waver->docnum){
    pthread_mutex_unlock(&g_mutex);
    return NULL;
  }
  while(queue_rnum(waver->queue) < 1 && g_thnum > 0){
    pthread_mutex_unlock(&g_mutex);
    est_usleep(SLEEPUSEC);
    if(pthread_mutex_lock(&g_mutex) != 0) return NULL;
  }
  rv = NULL;
  while(TRUE){
    if(!(buf = queue_dequeue(waver->queue))) break;
    if((rp = strchr(buf, '\t')) != NULL && (*dp = atoi(buf)) >= 0 &&
       vlvsiz(waver->trace, ++rp, -1) == -1){
      rv = cbmemdup(rp, -1);
      vlput(waver->trace, rv, -1, "", 0, VL_DKEEP);
      free(buf);
      if(waver->interval > 0) est_usleep(waver->interval * 1000);
      break;
    }
    free(buf);
  }
  pthread_mutex_unlock(&g_mutex);
  return rv;
}


/* add traced URLs */
static void enqueuelinks(WAVER *waver, CBLIST *links, CBMAP *kwords, int depth){
  const char *vbuf;
  char *ubuf, *pv, *tmp;
  int i, vsiz, lnum, slash, vnum, *svec, *tvec;
  double similarity, priority;
  if(depth >= waver->maxdepth) return;
  if(pthread_mutex_lock(&g_mutex) != 0) return;
  if(depth < 1 || cbmaprnum(waver->kwords) == 0){
    similarity = 1.0;
  } else {
    vnum = waver->keynum;
    svec = cbmalloc(vnum * sizeof(int));
    tvec = cbmalloc(vnum * sizeof(int));
    est_vector_set_seed(waver->kwords, svec, vnum);
    est_vector_set_target(waver->kwords, kwords, tvec, vnum);
    similarity = est_vector_cosine(svec, tvec, vnum) * 0.9 + 0.1;
    free(tvec);
    free(svec);
  }
  lnum = cblistnum(links) + 4;
  for(i = 0; i < cblistnum(links); i++){
    vbuf = cblistval(links, i, &vsiz);
    ubuf = cbmemdup(vbuf, vsiz);
    if((pv = strchr(ubuf, '#')) != NULL) *pv = '\0';
    if((waver->allowrx && !est_regex_match(waver->allowrx, ubuf)) ||
       (waver->denyrx && est_regex_match(waver->denyrx, ubuf)) ||
       vlvsiz(waver->trace, ubuf, -1) != -1){
      free(ubuf);
      continue;
    }
    slash = 6;
    for(pv = ubuf; *pv != '\0'; pv++){
      switch(*pv){
      case '/': slash += 1; break;
      case '?': slash += 5; break;
      case '&': slash += 1; break;
      case ';': slash += 1; break;
      }
    }
    priority = (similarity * 100) / (depth + 4) / pow(lnum, 0.7) / slash;
    priority *= (lnum - (i / 16.0)) / lnum;
    tmp = cbsprintf("%d\t%s", depth + 1, ubuf);
    queue_enqueue(waver->queue, tmp, 1.0 - priority);
    free(tmp);
    free(ubuf);
  }
  pthread_mutex_unlock(&g_mutex);
}


/* parse arguments of the init command */
static int runinit(int argc, char **argv){
  char *rootdir;
  int i, rv;
  rootdir = NULL;
  for(i = 2; i < argc; i++){
    if(!rootdir && argv[i][0] == '-'){
      usage();
    } else if(!rootdir){
      rootdir = argv[i];
    } else {
      usage();
    }
  }
  if(!rootdir) usage();
  rv = procinit(rootdir);
  return rv;
}


/* parse arguments of the crawl command */
static int runcrawl(int argc, char **argv){
  char *rootdir;
  int i, cont, rv;
  rootdir = NULL;
  cont = FALSE;
  for(i = 2; i < argc; i++){
    if(!rootdir && argv[i][0] == '-'){
      if(!strcmp(argv[i], "-cont")){
        cont = TRUE;
      } else {
        usage();
      }
    } else if(!rootdir){
      rootdir = argv[i];
    } else {
      usage();
    }
  }
  if(!rootdir) usage();
  rv = proccrawl(rootdir, cont);
  return rv;
}


/* parse arguments of the unittest command */
static int rununittest(int argc, char **argv){
  char *rootdir;
  int i, rv;
  rootdir = NULL;
  for(i = 2; i < argc; i++){
    if(!rootdir && argv[i][0] == '-'){
      usage();
    } else if(!rootdir){
      rootdir = argv[i];
    } else {
      usage();
    }
  }
  if(!rootdir) usage();
  rv = procunittest(rootdir);
  return rv;
}


/* parse arguments of the fetch command */
static int runfetch(int argc, char **argv){
  char *url;
  int i, rv;
  url = NULL;
  for(i = 2; i < argc; i++){
    if(!url && argv[i][0] == '-'){
      if(!strcmp(argv[i], "-proxy")){
        if(++i >= argc) usage();
        g_pxhost = argv[i];
        if(++i >= argc) usage();
        g_pxport = atoi(argv[i]);
      } else if(!strcmp(argv[i], "-tout")){
        if(++i >= argc) usage();
        g_timeout = atoi(argv[i]);
      } else if(!strcmp(argv[i], "-il")){
        if(++i >= argc) usage();
        g_inputlang = strtolang(argv[i]);
      } else {
        usage();
      }
    } else if(!url){
      url = argv[i];
    } else {
      usage();
    }
  }
  if(!url) usage();
  rv = procfetch(url);
  return rv;
}


/* perform the init command */
static int procinit(const char *rootdir){
  if(!waver_init(rootdir)){
    log_print(LL_ERROR, "initializing the root directory failed");
    return 1;
  }
  log_open(rootdir, LOGFILE, LL_INFO, FALSE);
  log_print(LL_INFO, "the root directory created");
  return 0;
}


/* perform the init command */
static int proccrawl(const char *rootdir, int cont){
  pthread_t th;
  TARGSURL *targs;
  WAVER *waver;
  const char *kbuf;
  char *tmp;
  int i, err, snum, depth;
  if(!(waver = waver_open(rootdir))){
    log_print(LL_ERROR, "%s: could not open", rootdir);
    return 1;
  }
  err = FALSE;
  log_print(LL_INFO, "crawling started %s", cont ? "(continue)" : "");
  if(!cont){
    while((tmp = queue_dequeue(waver->queue)) != NULL){
      free(tmp);
    }
    vlcurfirst(waver->trace);
    while(vlrnum(waver->trace) > 0){
      vlcurout(waver->trace);
    }
    cbmapiterinit(waver->seeds);
    while((kbuf = cbmapiternext(waver->seeds, NULL)) != NULL){
      tmp = cbsprintf("0\t%s", kbuf);
      queue_enqueue(waver->queue, tmp, 0.0);
      free(tmp);
    }
  }
  snum = queue_rnum(waver->queue);
  for(i = 0; !g_sigterm && (tmp = dequeue(waver, &depth)) != NULL; i++){
    targs = cbmalloc(sizeof(TARGSURL));
    targs->waver = waver;
    targs->url = tmp;
    targs->depth = depth;
    if(waver->thnum > 1){
      while(accessthnum(0) >= waver->thnum){
        est_usleep(SLEEPUSEC);
      }
      if(pthread_create(&th, NULL, geturldoc, targs) == 0){
        pthread_detach(th);
        if(i <= snum) est_usleep(SLEEPUSEC);
      } else {
        geturldoc(targs);
      }
    } else {
      geturldoc(targs);
    }
  }
  est_usleep(SLEEPUSEC);
  if(waver->thnum > 1){
    while(accessthnum(0) > 0){
      est_usleep(SLEEPUSEC);
    }
  }
  log_print(LL_INFO, "crawling finished");
  if(!waver_close(waver)){
    log_print(LL_ERROR, "%s: closing failed", rootdir);
    err = TRUE;
  }
  if(!err) log_print(LL_INFO, "finished successfully");
  return err ? 1 : 0;
}


/* perform the unittest command */
static int procunittest(const char *rootdir){
  WAVER *waver;
  QUEUE *queue;
  CBMAP *seeds, *kwords;
  char uri[URIBUFSIZ], *vbuf;
  int i, err;
  log_print(LL_INFO, "initializing the waver handle");
  if(!waver_init(rootdir)){
    log_print(LL_ERROR, "%s: initializing failed", rootdir);
    return FALSE;
  }
  log_print(LL_INFO, "opening the waver handle");
  if(!(waver = waver_open(rootdir))){
    log_print(LL_ERROR, "%s: opening failed", rootdir);
    return FALSE;
  }
  err = FALSE;
  log_print(LL_INFO, "checking seeding");
  seeds = waver->seeds;
  for(i = 0; i < 100; i++){
    sprintf(uri, "http://%05d/%x/%x.html",
            i + 1, (int)(est_random() * 0x1000000), (int)(est_random() * 0x1000000));
    cbmapput(seeds, uri, -1, "", 0, TRUE);
  }
  log_print(LL_INFO, "checking priority queue");
  queue = waver->queue;
  for(i = 0; i < 100; i++){
    sprintf(uri, "0\thttp://%05d/%x/%x.html",
            i + 1, (int)(est_random() * 0x1000000), (int)(est_random() * 0x1000000));
    if(!queue_enqueue(queue, uri, est_random())){
      err = TRUE;
      break;
    }
    if(i % 10 == 0) queue_set_range(queue, i);
  }
  if(err) log_print(LL_ERROR, "%s: enqueue failed", rootdir);
  for(i = 0; (vbuf = queue_dequeue(queue)) != NULL; i++){
    free(vbuf);
  }
  if(i != 100){
    err = TRUE;
    log_print(LL_ERROR, "%s: dequeue failed", rootdir);
  }
  log_print(LL_INFO, "checking keyword map");
  kwords = waver->kwords;
  for(i = 0; i < 10000; i++){
    sprintf(uri, "%d", (int)(est_random() * 1000));
    kwords_add(kwords, uri, (int)(est_random() * 1000));
  }
  kwords_reduce(kwords, 100);
  log_print(LL_INFO, "closing the waver handle");
  if(!waver_close(waver)){
    log_print(LL_ERROR, "%s: closing failed", rootdir);
    err = TRUE;
  }
  if(!err) log_print(LL_INFO, "finished successfully");
  return err ? 1 : 0;
}


/* perform the fetch command */
static int procfetch(const char *url){
  CBMAP *heads;
  CBLIST *links;
  CBDATUM *raw;
  ESTDOC *doc;
  const char *border, *vbuf;
  char *str;
  int i, vsiz;
  raw = cbdatumopen("", 0);
  heads = cbmapopen();
  links = cblistopen();
  doc = est_doc_new();
  if(!fetch_document(url, g_pxhost, g_pxport, g_timeout, NULL, raw, heads,
                     links, doc, g_inputlang)){
    est_doc_delete(doc);
    cblistclose(links);
    cbmapclose(heads);
    cbdatumclose(raw);
    return 1;
  }
  border = est_border_str();
  printf("URL: %s\r\n", url);
  str = cbdatestrhttp(time(NULL), 0);
  printf("Date: %s\r\n", str);
  free(str);
  str = cbmimeencode((vbuf = est_doc_attr(doc, ESTDATTRTITLE)) ? vbuf : url, "UTF-8", TRUE);
  printf("Subject: [estwaver] %s\r\n", str);
  free(str);
  printf("Content-Type: multipart/mixed; boundary=%s\r\n", border);
  printf("\r\n");
  printf("This is a multi-part message in MIME format.\n");
  printf("\r\n");
  printf("--%s\r\n", border);
  printf("Content-Type: text/x-estraier-draft\r\n");
  printf("X-Estwaver-Role: draft\r\n");
  printf("\r\n");
  str = est_doc_dump_draft(doc);
  printf("%s", str);
  free(str);
  printf("\r\n");
  printf("--%s\r\n", border);
  printf("Content-Type: text/plain\r\n");
  printf("X-Estwaver-Role: links\r\n");
  printf("\r\n");
  for(i = 0; i < cblistnum(links); i++){
    printf("%s\n", cblistval(links, i, NULL));
  }
  printf("\r\n");
  printf("--%s\r\n", border);
  cbmapiterinit(heads);
  while((vbuf = cbmapiternext(heads, &vsiz)) != NULL){
    if(vsiz < 1){
      printf("X-Original-HTTP-Response: %s\r\n", cbmapget(heads, vbuf, vsiz, NULL));
    } else if(!strcmp(vbuf, "content-encoding")){
      printf("X-Original-Content-Encoding: %s\r\n", cbmapget(heads, vbuf, vsiz, NULL));
    } else {
      str = capitalize(vbuf);
      printf("%s: %s\r\n", str, cbmapget(heads, vbuf, vsiz, NULL));
      free(str);
    }
  }
  printf("X-Estwaver-Role: raw\r\n");
  printf("\r\n");
  fwrite(cbdatumptr(raw), 1, cbdatumsize(raw), stdout);
  printf("\r\n");
  printf("--%s--\r\n", border);
  est_doc_delete(doc);
  cblistclose(links);
  cbmapclose(heads);
  cbdatumclose(raw);
  return 0;
}


/* get the language value */
static int strtolang(const char *str){
  if(!cbstricmp(str, "en")) return ESTLANGEN;
  if(!cbstricmp(str, "ja")) return ESTLANGJA;
  if(!cbstricmp(str, "zh")) return ESTLANGZH;
  if(!cbstricmp(str, "ko")) return ESTLANGKO;
  return ESTLANGMISC;
}


/* make a capitalized string */
static char *capitalize(const char *str){
  char *buf;
  int i, cap;
  buf = cbmemdup(str, -1);
  cap = TRUE;
  for(i = 0; buf[i] != '\0'; i++){
    if(cap && buf[i] >= 'a' && buf[i] <= 'z'){
      buf[i] -= 'a' - 'A';
    }
    cap = buf[i] == ' ' || buf[i] == '-';
  }
  return buf;
}


/* access the number of threads */
static int accessthnum(int inc){
  int thnum;
  if(pthread_mutex_lock(&g_mutex) != 0) return INT_MAX;
  g_thnum += inc;
  thnum = g_thnum;
  pthread_mutex_unlock(&g_mutex);
  return thnum;
}


/* get a document of URL */
static void *geturldoc(void *args){
  TARGSURL *myargs;
  WAVER *waver;
  CBMAP *heads, *kwords;
  CBLIST *links;
  CBDATUM *raw;
  ESTDOC *doc;
  const char *kbuf, *vbuf;
  int ksiz, num;
  double bias;
  accessthnum(1);
  myargs = args;
  waver = myargs->waver;
  raw = cbdatumopen("", 0);
  heads = cbmapopen();
  links = cblistopen();
  doc = est_doc_new();
  log_print(LL_INFO, "fetching: %d: %s", myargs->depth, myargs->url);
  if(fetch_document(myargs->url, waver->pxhost, waver->pxport, waver->timeout,
                    waver->mtrules, raw, heads, links, doc, waver->language) &&
     est_doc_attr(doc, ESTDATTRURI) && !est_doc_is_empty(doc)){
    log_print(LL_DEBUG, "registering: %s", myargs->url);
    est_doc_slim(doc, waver->textlimit);
    if(pthread_mutex_lock(&g_mutex) == 0){
      kwords = est_morph_etch_doc(doc, waver->keynum);
      pthread_mutex_unlock(&g_mutex);
    } else {
      kwords = est_mtdb_etch_doc(NULL, doc, waver->keynum);
    }
    if((vbuf = cbmapget(waver->seeds, myargs->url, -1, NULL)) != NULL){
      if((bias = strtod(vbuf, NULL)) > 0.0){
        cbmapiterinit(kwords);
        while((kbuf = cbmapiternext(kwords, &ksiz)) != NULL){
          num = atoi(cbmapget(kwords, kbuf, ksiz, NULL)) * bias;
          if(num > 0) kwords_add(waver->kwords, kbuf, num);
        }
      }
    }
    if(!est_mtdb_put_doc(waver->index, doc, ESTPDCLEAN) ||
       !est_mtdb_put_keywords(waver->index, est_doc_id(doc), kwords))
      log_print(LL_ERROR, "registration failed: %s: %s", myargs->url, dperrmsg(dpecode));
    enqueuelinks(waver, links, kwords, myargs->depth);
    cbmapclose(kwords);
  } else {
    log_print(LL_WARN, "ignored: %s", myargs->url);
  }
  est_doc_delete(doc);
  cblistclose(links);
  cbmapclose(heads);
  cbdatumclose(raw);
  free(myargs->url);
  free(myargs);
  accessthnum(-1);
  return NULL;
}



/* END OF FILE */
